
(* print_string is just a pre-defined function (not a keyword or anything) *)
let x = print_string "Hello, World!\n"

(* underscore is not a variable *)
let _ = print_string "Hello, World!\n"

let hello_str = "Hello, World!\n"

(* can refer to previously declared variables *)
(* function declarations look a lot like variable declarations *)
let f x = print_string hello_str

(* x has already been evaluated to () *)
let _ = x

(* function application is juxtaposition -- get used to it *)
(* f is polymorphic -- degenerately because it ignores x   *)
let _ = f 37
let _ = f f

(* Every expression has one type -- we've seen int, string, unit,
   t1 -> t2 and type variables 'a *)

(* an explicit type can help with debugging or to artificially constrain *)
let g (x : int) = f x

let _ = g 37
(*let _ = g "hi mom"*)

(* A really small infinite loop -- do not worry about the inferred type *)
(* rec is necesary -- else forever is unbound (would refer to earlier one) *)
let rec forever x = forever x

(* A more useful recursive function, incorrect for negative x *)
(* Note precedence requires parentheses here *)
let rec fact x = if x == 0 then 1 else x * (fact (x - 1))

(* parentheses necessary here too -- see currying below *)
(* ^ is string concatenation *)
let _ = print_string ((string_of_int (fact 6)) ^ "\n")

(* everything is an expression *)
let rec fact2 x = (if x == 0 then 1 else x * fact (x - 1)) * 2 / 2

(* local variables and functions look very similar to top-level ones *)
(* indentation is just style *)
let quadruple x =
    let double y = y + y in
    let ans = double x + double x in
    ans

let _ = print_string ((string_of_int (quadruple 7)) ^ "\n")

(* functions can be anonymous *)
let quadruple2 x = (fun x -> x + x) x + (fun x -> x + x) x

let _ = print_string ((string_of_int (quadruple2 7)) ^ "\n")

(* In fact, we've been using a shortcut:
   "let f x = e" is really just "let f = fun x -> e"
   (Show with quadruple.)
*)

(* Higher-order programming is great for reuse *)
(* The boring way... *)
let print_int_and_newline i = print_string ((string_of_int i) ^ "\n")

(* With passing functions... (and our first 2-argument function) *)
(* and useful polymorphism -- though we're not using it *)
let print_int_and_newline2 f i = print_int_and_newline (f i)
let _ = print_int_and_newline (fact 6)
let _ = print_int_and_newline (quadruple 7)
let _ = print_int_and_newline2 fact 6
let _ = print_int_and_newline2 quadruple 7

(* inferior style... *)
let print_factof_and_newline1 i = print_int_and_newline2 fact i
let print_quadrupleof_and_newline1 i = print_int_and_newline2 quadruple i

(* The camlers delight -- partial application *)
let print_factof_and_newline = print_int_and_newline2 fact

let print_quadrupleof_and_newline = print_int_and_newline2 quadruple

let _ = print_factof_and_newline1 6
let _ = print_quadrupleof_and_newline1 7
let _ = print_factof_and_newline 6
let _ = print_quadrupleof_and_newline 7

(* What's really going on:
   * Every function takes exactly one argument
   * "Multiple arguments is syntactic sugar"
   * Application "associates to the left"
 *)
let print_int_and_newline2 = fun f -> (fun i -> print_int_and_newline (f i))

let _ = (print_int_and_newline2 fact) 6

(* closures and static scoping *)
let do_fact6 =
   let x = 6 in
   fun () -> fact x

let x = 7
let _ = print_string "demonstrate closures..."
let _ = print_int_and_newline (do_fact6 ())

(* Variables are not mutable *)
let x = 42
let f() = x
let x = x + 1
let _ = print_int_and_newline (f())


(* Refs are mutable, but should be used sparingly. *)
let x = ref 42
let f () = !x
let _ = print_int_and_newline (f())
let _ = x := !x + 1
let _ = print_int_and_newline (f())

(************************ Aggregate Types ******************************)

(* records -- definition, creation, projection *)
type int_pair = { first : int; second : int}

let pair1 = { first = 3; second = 4}

let sum_int_pair x = x.first + x.second

let sum = sum_int_pair pair1 + sum_int_pair {first = 1; second = 2}

(* all record labels must be distinct (namespaces make this a bit less annoying)
   the reason is type inference -- see sum_int_pair 
   polymorphic type constructurs
*)
type 'a pair = { a_first : 'a; a_second : 'a }

let sum_a_pair f x = f x.a_first + f x.a_second

let str_pair = {a_first = "hi"; a_second = "mom"}
let int_pair = {a_first = 3;    a_second = 4} (*shadows type-name -- bad style*)
let int_pair_pair = {a_first = int_pair; a_second = int_pair}
let sum_str = sum_a_pair (fun x -> 1) str_pair
let sum_int = sum_a_pair (fun x -> x) int_pair
let _ = print_int_and_newline sum_str
let _ = print_int_and_newline sum_int

(* discriminated unions also known as "dataypes" -- definition, creation,
   proection *)
type foo = Foo of int | Bar of int_pair | Baz of int * int | Quux

let foo3 = Foo (1 + 2)
let bar12 = Bar pair1
let baz1_120 = Baz(1,fact 5)
let quux = Quux (* not much point in this one *)

let is_a_foo x =
  match x with 
   Foo i -> true
 | Bar pr -> false
 | Baz(i,j) -> false
 | Quux -> false

let sum_foo x =
  match x with
    Foo i -> i
  | Bar pr -> sum_int_pair pr
  | Baz(i,j) -> i + j
  | Quux -> 0
(* Note: variant names must be capitalized, other identifiers must not be. *)

(* recursive types *)
type 'a lst = Null | Cons of 'a * 'a lst

let lst1 = Cons(3,Null)
(* let lst_bad = Cons("hi",lst1) *)
let lst2 = Cons(1,Cons(2,lst1))

let rec len l =
  match l with
    Null -> 0
  | Cons(x,rest) -> 1 + len rest

let rec sum l =
  match l with
    Null -> 0
  | Cons(x,rest) -> x + sum rest

let nest = Cons (Cons (3,Null), Cons (Cons (4,Null), Null))

(* In fact 'a list is built-in and has funny syntax:
   [] is like Null
   :: (infix) is like Cons *)

let list1 = 3::[]
let list2 = 1::2::list1
let list3 = [5;6;7]

let rec len2 l =
  match l with
    [] -> 0
  | x::rest -> 1 + len2 rest

let rec sum2 l =
  match l with
    [] -> 0
  | x::rest -> x + sum2 rest

(* Moreover, booleans and if are built-in funny syntax (uncapitalized) *)

let print_bool1 b = 
  print_string (if b then "true\n" else "false\n")
let print_bool2 b = 
  print_string (match b with true -> "true\n" | false -> "false\n")

let _ = print_bool1 true; print_bool2 false

(* tuples -- defining record types all the time is a pain and unnecessary *)
let x = (3,"hi",(fun x -> x), fun x -> x ^ "ism")

let z =
  match x with
    (i,s,f1,f2) -> f1 i
(* nasty error message without parens around identity function *)
(* also (i,_,f1,_).  also let expressions *)

type s = Pos | Neg

let multsign x1 x2 =
  let sign x = if(x >= 0) then Pos else Neg in
  match(sign x1, sign x2) with
    (Pos,Pos) -> Pos
  | (Neg,Neg) -> Pos
  | _ -> Neg (* dangerous and unnecessary, but convenient *)

(* exceptions *)
exception MyException
let mydiv x y =
  if y=0 then raise MyException else x/y

let yourdiv x y e =
  if y=0 then raise e else x/y

(* exceptions can carry values too -- see the manual *)
